function stringifyQuery(api) {
    // Each key in query, if it's an object, will be stringified
    for (const key of Object.keys(api.query)) {
        if (typeof api.query[key] === 'object')
            api.query[key] = JSON.stringify(api.query[key]);
    }
}
export default function handler(options, api) {
    const { fetch } = options;
    const stringifyQueryObjects = options.config.stringifyQueryObjects === undefined ? true : options.config.stringifyQueryObjects;
    if (stringifyQueryObjects && 'query' in api)
        stringifyQuery(api);
    const url = options.config.baseUrl || '' + options.path + ('query' in api ? '?' + new URLSearchParams(api.query).toString() : '');
    const baseData = options.config.baseData || {};
    const isForm = Object.prototype.toString.call(api.body) === '[object FormData]';
    if (!('content-type' in api.headers))
        api.headers['content-type'] = api.headers['content-type'] || isForm ? 'multipart/form-data' : 'application/json';
    api.body = 'body' in api ? typeof api.body === 'object' && JSON.stringify(api.body) : '{}';
    if (api.method === 'GET' && !('cache-control' in api.headers))
        api.headers['cache-control'] = 'public, max-age=604800, immutable';
    api.headers['x-requested-with'] = 'sveltekit-zero-api';
    const callbacks = [];
    // * $. callback handlers
    const $ = {
        callback: undefined,
        promise: undefined,
        resolve: undefined,
    };
    // Resolves the whole request
    let resolver;
    let promiser = new Promise(resolve => resolver = resolve);
    let fetchAPI = {
        ...callbackHandler((s, cb) => {
            callbacks.push([s, cb]);
            return proxy;
        }),
        ...promiser
    };
    fetchAPI['_'] = fetchAPI;
    fetchAPI['$'] = callbackHandler((s, cb) => {
        $.callback = [s, cb];
        $.promise = new Promise(resolve => $.resolve = resolve);
        $.promise.then(d => resolver(d));
        return promiser;
    });
    const response = fetch(url, { ...baseData, ...api, headers: { ...(baseData['headers'] || {}), ...(api['headers'] || {}) } });
    response.then(async (res) => {
        const json = (res.headers.get('content-type') || '').includes('application/json') && await res[options.config.format || 'json']();
        // TODO: Handle other responses than just JSON
        const response = {
            body: json || res.body,
            bodyUsed: res.bodyUsed,
            clone: res.clone,
            headers: res.headers,
            ok: res.ok,
            redirected: res.redirected,
            status: res.status,
            statusText: res.statusText,
            type: res.type,
            url: res.url
        };
        if (!res.bodyUsed) {
            response.blob = res.blob;
            response.arrayBuffer = res.arrayBuffer;
            response.formData = res.formData;
            response.text = res.text;
            response.json = res.json;
        }
        respond(callbacks, response, options, $);
        if (!$.promise)
            resolver(response);
    });
    response.catch((err) => {
        if ('onError' in options.config)
            options.config.onError && options.config.onError(err);
        throw err;
    });
    const proxy = new Proxy(fetchAPI, {
        get(target, prop) {
            if (prop === 'then')
                return promiser.then.bind(promiser);
            if (prop === 'catch')
                return promiser.catch.bind(promiser);
            if (prop === 'finally')
                return promiser.finally.bind(promiser);
            return target[prop];
        }
    });
    return proxy;
}
function respond(callbacks, res, options, $) {
    let statusRange = Math.floor(res.status / 100);
    let responseError = statusRange === 4 || statusRange === 5;
    const validStatus = (code) => code === 0 || code === statusRange || code === res.status || (responseError && code === 45);
    const { prependCallbacks } = options.config;
    if (prependCallbacks) {
        let cbs = [];
        let prepender = callbackHandler((s, cb) => {
            cbs.push([s, cb]);
            return prepender;
        });
        prependCallbacks(prepender);
        callbacks = [...cbs, ...callbacks];
    }
    callbacks
        .filter(([code]) => validStatus(code))
        .forEach(([code, cb]) => { cb(res); });
    if ($.resolve && $.callback) {
        if (validStatus($.callback[0]))
            $.resolve($.callback[1](res));
        else
            $.resolve(undefined);
    }
    if ('onSuccess' in options.config)
        options.config.onSuccess && options.config.onSuccess(res);
}
function callbackHandler(handleCallback) {
    return {
        // "Any response"
        Any: function (cb) { return handleCallback(0, cb); },
        // general
        Informational: function (cb) { return handleCallback(1, cb); },
        Success: function (cb) { return handleCallback(2, cb); },
        Redirect: function (cb) { return handleCallback(3, cb); },
        Error: function (cb) { return handleCallback(45, cb); },
        ClientError: function (cb) { return handleCallback(4, cb); },
        ServerError: function (cb) { return handleCallback(5, cb); },
        // 1××
        Continue: function (cb) { return handleCallback(100, cb); },
        SwitchingProtocols: function (cb) { return handleCallback(101, cb); },
        Processing: function (cb) { return handleCallback(102, cb); },
        EarlyHints: function (cb) { return handleCallback(103, cb); },
        // 2××
        Ok: function (cb) { return handleCallback(200, cb); },
        Created: function (cb) { return handleCallback(201, cb); },
        Accepted: function (cb) { return handleCallback(202, cb); },
        NonAuthoritativeInformation: function (cb) { return handleCallback(203, cb); },
        NoContent: function (cb) { return handleCallback(204, cb); },
        ResetContent: function (cb) { return handleCallback(205, cb); },
        PartialContent: function (cb) { return handleCallback(206, cb); },
        MultiStatus: function (cb) { return handleCallback(207, cb); },
        AlreadyReported: function (cb) { return handleCallback(208, cb); },
        IMUsed: function (cb) { return handleCallback(226, cb); },
        // 3××
        MultipleChoices: function (cb) { return handleCallback(300, cb); },
        MovedPermanently: function (cb) { return handleCallback(301, cb); },
        Found: function (cb) { return handleCallback(302, cb); },
        SeeOther: function (cb) { return handleCallback(303, cb); },
        NotModified: function (cb) { return handleCallback(304, cb); },
        UseProxy: function (cb) { return handleCallback(305, cb); },
        SwitchProxy: function (cb) { return handleCallback(306, cb); },
        TemporaryRedirect: function (cb) { return handleCallback(307, cb); },
        PermanentRedirect: function (cb) { return handleCallback(308, cb); },
        // 4××
        BadRequest: function (cb) { return handleCallback(400, cb); },
        Unauthorized: function (cb) { return handleCallback(401, cb); },
        PaymentRequired: function (cb) { return handleCallback(402, cb); },
        Forbidden: function (cb) { return handleCallback(403, cb); },
        NotFound: function (cb) { return handleCallback(404, cb); },
        MethodNotAllowed: function (cb) { return handleCallback(405, cb); },
        NotAcceptable: function (cb) { return handleCallback(406, cb); },
        ProxyAuthenticationRequired: function (cb) { return handleCallback(407, cb); },
        RequestTimeout: function (cb) { return handleCallback(408, cb); },
        Conflict: function (cb) { return handleCallback(409, cb); },
        Gone: function (cb) { return handleCallback(410, cb); },
        LengthRequired: function (cb) { return handleCallback(411, cb); },
        PreconditionFailed: function (cb) { return handleCallback(412, cb); },
        PayloadTooLarge: function (cb) { return handleCallback(413, cb); },
        URITooLong: function (cb) { return handleCallback(414, cb); },
        UnsupportedMediaType: function (cb) { return handleCallback(415, cb); },
        RangeNotSatisfiable: function (cb) { return handleCallback(416, cb); },
        ExpectationFailed: function (cb) { return handleCallback(417, cb); },
        ImATeapot: function (cb) { return handleCallback(418, cb); },
        MisdirectedRequest: function (cb) { return handleCallback(421, cb); },
        UnprocessableEntity: function (cb) { return handleCallback(422, cb); },
        Locked: function (cb) { return handleCallback(423, cb); },
        FailedDependency: function (cb) { return handleCallback(424, cb); },
        TooEarly: function (cb) { return handleCallback(425, cb); },
        UpgradeRequired: function (cb) { return handleCallback(426, cb); },
        PreconditionRequired: function (cb) { return handleCallback(428, cb); },
        TooManyRequests: function (cb) { return handleCallback(429, cb); },
        RequestHeaderFieldsTooLarge: function (cb) { return handleCallback(431, cb); },
        UnavailableForLegalReasons: function (cb) { return handleCallback(451, cb); },
        // 5××
        InternalServerError: function (cb) { return handleCallback(500, cb); },
        NotImplemented: function (cb) { return handleCallback(501, cb); },
        BadGateway: function (cb) { return handleCallback(502, cb); },
        ServiceUnavailable: function (cb) { return handleCallback(503, cb); },
        GatewayTimeout: function (cb) { return handleCallback(504, cb); },
        HTTPVersionNotSupported: function (cb) { return handleCallback(505, cb); },
        VariantAlsoNegotiates: function (cb) { return handleCallback(506, cb); },
        InsufficientStorage: function (cb) { return handleCallback(507, cb); },
        LoopDetected: function (cb) { return handleCallback(508, cb); },
        NotExtended: function (cb) { return handleCallback(510, cb); },
        NetworkAuthenticationRequired: function (cb) { return handleCallback(511, cb); }
    };
}
